"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Option_1 = require("fp-ts/lib/Option");
var Reader_1 = require("fp-ts/lib/Reader");
var Array_1 = require("fp-ts/lib/Array");
var ts = require("typescript");
var M = require("./model");
var monocle_ts_1 = require("monocle-ts");
exports.defaultOptions = {
    tagName: 'type',
    foldName: 'fold',
    matcheeName: 'fa',
    handlersStyle: { type: 'positional' },
    encoding: 'literal'
};
var getLens = monocle_ts_1.Lens.fromProp();
exports.lenses = {
    tagName: getLens('tagName'),
    foldName: getLens('foldName'),
    matcheeName: getLens('matcheeName'),
    handlersStyle: getLens('handlersStyle'),
    encoding: getLens('encoding')
};
var getMemberName = function (m, position) {
    return m.name.getOrElseL(function () { return "value" + position; });
};
var getDomainParameterName = function (type) {
    switch (type.kind) {
        case 'Ref':
            return Option_1.some(getFirstLetterLowerCase(type.name));
        case 'Tuple':
            return Option_1.some('tuple');
        case 'Fun':
            return Option_1.some('f');
        case 'Unit':
            return Option_1.none;
    }
};
var getTypeNode = function (type) {
    switch (type.kind) {
        case 'Ref':
            return ts.createTypeReferenceNode(type.name, type.parameters.map(function (p) { return getTypeNode(p); }));
        case 'Tuple':
            return ts.createTupleTypeNode(type.types.map(getTypeNode));
        case 'Fun':
            return ts.createFunctionTypeNode(undefined, getDomainParameterName(type.domain)
                .map(function (domainName) { return [getParameterDeclaration(domainName, getTypeNode(type.domain))]; })
                .getOrElse([]), getTypeNode(type.codomain));
        case 'Unit':
            return ts.createTypeReferenceNode('undefined', []);
    }
};
var getDataLiteralEncoding = function (d) {
    return new Reader_1.Reader(function (e) {
        var unionType = ts.createUnionTypeNode(d.constructors.toArray().map(function (c) {
            var tag = getPropertySignature(e.tagName, ts.createLiteralTypeNode(ts.createLiteral(c.name)));
            var members = c.members.map(function (m, position) {
                return getPropertySignature(getMemberName(m, position), getTypeNode(m.type));
            });
            return ts.createTypeLiteralNode([tag].concat(members));
        }));
        return [
            getTypeAliasDeclaration(d.name, unionType, d.parameterDeclarations.map(function (p) {
                return ts.createTypeParameterDeclaration(p.name, p.constraint.map(getTypeNode).toUndefined());
            }))
        ];
    });
};
var getTypeAliasDeclaration = function (name, type, typeParameters) {
    return ts.createTypeAliasDeclaration(undefined, [ts.createModifier(ts.SyntaxKind.ExportKeyword)], name, typeParameters, type);
};
var URI2HKTNames = {
    1: 'URI2HKT',
    2: 'URI2HKT2',
    3: 'URI2HKT3',
    4: 'URI2HKT4'
};
var URI2HKTParametersNames = ['A', 'L', 'U', 'X'];
var getPropertySignature = function (name, type, isReadonly) {
    if (isReadonly === void 0) { isReadonly = true; }
    return ts.createPropertySignature(isReadonly ? [ts.createModifier(ts.SyntaxKind.ReadonlyKeyword)] : undefined, name, undefined, type, undefined);
};
var getInterfaceDeclaration = function (name, typeParameters, members) {
    return ts.createInterfaceDeclaration(undefined, undefined, name, typeParameters, undefined, members);
};
var getDataModuleDeclaration = function (d) {
    var len = getDataParametersLength(d);
    var URI2HKT = URI2HKTNames[len];
    var parameters = URI2HKTParametersNames.slice(0, len).reverse();
    return ts.createModuleDeclaration(undefined, [ts.createModifier(ts.SyntaxKind.DeclareKeyword)], ts.createStringLiteral('fp-ts/lib/HKT'), ts.createModuleBlock([
        getInterfaceDeclaration(URI2HKT, parameters.map(function (p) { return ts.createTypeParameterDeclaration(p); }), [
            getPropertySignature(d.name, ts.createTypeReferenceNode(d.name, parameters.map(function (p) { return ts.createTypeReferenceNode(p, []); })), false)
        ])
    ]), undefined);
};
var getPropertyDeclaration = function (name, type, inizializer, isExclamation) {
    if (isExclamation === void 0) { isExclamation = false; }
    return ts.createProperty(undefined, [ts.createModifier(ts.SyntaxKind.ReadonlyKeyword)], name, isExclamation ? ts.createToken(ts.SyntaxKind.ExclamationToken) : undefined, type, inizializer);
};
var getMethod = function (name, typeParameters, parameters, returnType, body) {
    return ts.createMethod(undefined, undefined, undefined, name, undefined, typeParameters, parameters, returnType, body);
};
var getFoldMethod = function (d, c, name, isEager) {
    var returnTypeParameterName = getFoldReturnTypeParameterName(d);
    var handlerExpression = ts.createIdentifier(getFoldHandlerName(c));
    var returnExpression = isNullaryConstructor(c) && isEager
        ? handlerExpression
        : ts.createCall(handlerExpression, undefined, c.members.map(function (m, position) {
            return ts.createPropertyAccess(ts.createThis(), getMemberName(m, position));
        }));
    var fold = getMethod(name, [ts.createTypeParameterDeclaration(returnTypeParameterName)], getFoldPositionalHandlers(d, isEager, c), ts.createTypeReferenceNode(returnTypeParameterName, []), ts.createBlock([ts.createReturn(returnExpression)]));
    return fold;
};
var getDataFptsEncoding = function (d) {
    return new Reader_1.Reader(function (e) {
        var unionType = ts.createUnionTypeNode(d.constructors.toArray().map(function (c) {
            return ts.createTypeReferenceNode(c.name, d.parameterDeclarations.map(function (p) { return ts.createTypeReferenceNode(p.name, []); }));
        }));
        var moduleDeclaration = getDataModuleDeclaration(d);
        var uriValue = getConstantDeclaration('URI', ts.createStringLiteral(d.name));
        var uriType = getTypeAliasDeclaration('URI', ts.createTypeQueryNode(ts.createIdentifier('URI')));
        var classes = d.constructors.toArray().map(function (c) {
            var typeParameters = d.parameterDeclarations.map(function (p) { return ts.createTypeParameterDeclaration(p.name); });
            var parameters = c.members.map(function (m, position) {
                return getParameterDeclaration(getMemberName(m, position), getTypeNode(m.type), true);
            });
            var isNullary = isNullaryConstructor(c);
            var constructor = ts.createConstructor(undefined, isNullary ? [ts.createModifier(ts.SyntaxKind.PrivateKeyword)] : undefined, parameters, ts.createBlock([]));
            var tag = getPropertyDeclaration('_tag', ts.createLiteralTypeNode(ts.createStringLiteral(c.name)), ts.createStringLiteral(c.name));
            var len = getDataParametersLength(d);
            var fptsParameters = URI2HKTParametersNames.slice(0, len).map(function (name) {
                return getPropertyDeclaration("_" + name, ts.createTypeReferenceNode(name, []), undefined, true);
            });
            var URI = getPropertyDeclaration('_URI', ts.createTypeReferenceNode('URI', []), undefined, true);
            var folds = Array_1.empty;
            if (isSumType(d)) {
                if (isEagerFoldSupported(d)) {
                    folds = [getFoldMethod(d, c, e.foldName, true), getFoldMethod(d, c, e.foldName + "L", false)];
                }
                else {
                    folds = [getFoldMethod(d, c, e.foldName, false)];
                }
            }
            var members = [tag].concat(fptsParameters, [URI, constructor], folds);
            if (isNullary) {
                members.unshift(ts.createProperty(undefined, [ts.createModifier(ts.SyntaxKind.StaticKeyword)], 'value', undefined, ts.createTypeReferenceNode(d.name, d.parameterDeclarations.map(function (_) { return ts.createTypeReferenceNode('never', []); })), ts.createNew(ts.createIdentifier(c.name), undefined, [])));
            }
            return ts.createClassDeclaration(undefined, [ts.createModifier(ts.SyntaxKind.ExportKeyword)], c.name, typeParameters, undefined, members);
        });
        return [
            moduleDeclaration,
            uriValue,
            uriType,
            getTypeAliasDeclaration(d.name, unionType, d.parameterDeclarations.map(function (p) {
                return ts.createTypeParameterDeclaration(p.name, p.constraint.map(getTypeNode).toUndefined());
            }))
        ].concat(classes);
    });
};
var getDataParametersLength = function (d) {
    return d.parameterDeclarations.length;
};
var shouldUseLiteralEncoding = function (d) {
    return new Reader_1.Reader(function (e) { return e.encoding === 'literal' || getDataParametersLength(d) > URI2HKTParametersNames.length; });
};
exports.data = function (d) {
    return shouldUseLiteralEncoding(d).chain(function (yes) { return (yes ? getDataLiteralEncoding(d) : getDataFptsEncoding(d)); });
};
var getFirstLetterLowerCase = function (name) {
    return name.substring(0, 1).toLocaleLowerCase() + name.substring(1);
};
var getConstantDeclaration = function (name, initializer, type) {
    return ts.createVariableStatement([ts.createModifier(ts.SyntaxKind.ExportKeyword)], ts.createVariableDeclarationList([ts.createVariableDeclaration(name, type, initializer)], ts.NodeFlags.Const));
};
var getFunctionDeclaration = function (name, typeParameters, parameters, type, body) {
    return ts.createFunctionDeclaration(undefined, [ts.createModifier(ts.SyntaxKind.ExportKeyword)], undefined, name, typeParameters, parameters, type, body);
};
var getParameterDeclaration = function (name, type, isReadonly) {
    if (isReadonly === void 0) { isReadonly = false; }
    return ts.createParameter(undefined, isReadonly ? [ts.createModifier(ts.SyntaxKind.ReadonlyKeyword)] : undefined, undefined, name, undefined, type, undefined);
};
var getLiteralNullaryConstructor = function (c, d) {
    return new Reader_1.Reader(function (e) {
        var name = getFirstLetterLowerCase(c.name);
        var type = ts.createTypeReferenceNode(d.name, d.parameterDeclarations.map(function (p) {
            return p.constraint.map(function (c) { return getTypeNode(c); }).getOrElse(ts.createKeywordTypeNode(ts.SyntaxKind.NeverKeyword));
        }));
        var initializer = ts.createObjectLiteral([ts.createPropertyAssignment(e.tagName, ts.createStringLiteral(c.name))]);
        return getConstantDeclaration(name, initializer, type);
    });
};
var getLiteralConstructor = function (c, d) {
    return new Reader_1.Reader(function (e) {
        var name = getFirstLetterLowerCase(c.name);
        var typeParameters = d.parameterDeclarations.map(function (p) {
            return ts.createTypeParameterDeclaration(p.name, p.constraint.map(getTypeNode).toUndefined());
        });
        var parameters = c.members.map(function (m, position) {
            var name = getMemberName(m, position);
            var type = getTypeNode(m.type);
            return getParameterDeclaration(name, type);
        });
        var type = ts.createTypeReferenceNode(d.name, d.parameterDeclarations.map(function (p) { return ts.createTypeReferenceNode(p.name, []); }));
        var body = ts.createBlock([
            ts.createReturn(ts.createObjectLiteral([
                ts.createPropertyAssignment(e.tagName, ts.createStringLiteral(c.name))
            ].concat(c.members.map(function (m, position) {
                var name = getMemberName(m, position);
                return ts.createShorthandPropertyAssignment(name);
            }))))
        ]);
        return getFunctionDeclaration(name, typeParameters, parameters, type, body);
    });
};
var getConstructorsLiteralEncoding = function (d) {
    var constructors = d.constructors.toArray().map(function (c) {
        if (c.members.length === 0) {
            return getLiteralNullaryConstructor(c, d);
        }
        else {
            return getLiteralConstructor(c, d);
        }
    });
    return Array_1.array.sequence(Reader_1.reader)(constructors);
};
var getFptsNullaryConstructor = function (c, d) {
    return new Reader_1.Reader(function (_) {
        var name = getFirstLetterLowerCase(c.name);
        var type = ts.createTypeReferenceNode(d.name, d.parameterDeclarations.map(function (p) {
            return p.constraint.map(function (c) { return getTypeNode(c); }).getOrElse(ts.createKeywordTypeNode(ts.SyntaxKind.NeverKeyword));
        }));
        var initializer = ts.createPropertyAccess(ts.createIdentifier(c.name), 'value');
        return getConstantDeclaration(name, initializer, type);
    });
};
var getFptsConstructor = function (c, d) {
    return new Reader_1.Reader(function (_) {
        var name = getFirstLetterLowerCase(c.name);
        var typeParameters = d.parameterDeclarations.map(function (p) {
            return ts.createTypeParameterDeclaration(p.name, p.constraint.map(getTypeNode).toUndefined());
        });
        var parameters = c.members.map(function (m, position) {
            var name = getMemberName(m, position);
            var type = getTypeNode(m.type);
            return getParameterDeclaration(name, type);
        });
        var type = ts.createTypeReferenceNode(d.name, d.parameterDeclarations.map(function (p) { return ts.createTypeReferenceNode(p.name, []); }));
        var body = ts.createBlock([
            ts.createReturn(ts.createNew(ts.createIdentifier(c.name), undefined, c.members.map(function (m, position) {
                return ts.createIdentifier(getMemberName(m, position));
            })))
        ]);
        return getFunctionDeclaration(name, typeParameters, parameters, type, body);
    });
};
var getConstructorsFptsEncoding = function (d) {
    var constructors = d.constructors.toArray().map(function (c) {
        if (c.members.length === 0) {
            return getFptsNullaryConstructor(c, d);
        }
        else {
            return getFptsConstructor(c, d);
        }
    });
    return Array_1.array.sequence(Reader_1.reader)(constructors);
};
exports.constructors = function (d) {
    return shouldUseLiteralEncoding(d).chain(function (yes) {
        return yes ? getConstructorsLiteralEncoding(d) : getConstructorsFptsEncoding(d);
    });
};
var isNullaryConstructor = function (c) {
    return c.members.length === 0;
};
var isEagerFoldSupported = function (d) {
    return d.constructors.toArray().some(isNullaryConstructor);
};
var isSumType = function (d) {
    return d.constructors.length() > 1;
};
var getFoldReturnTypeParameterName = function (d) {
    var base = 'R';
    var candidate = base;
    var counter = 0;
    while (d.parameterDeclarations.findIndex(function (_a) {
        var name = _a.name;
        return candidate === name;
    }) !== -1) {
        candidate = base + ++counter;
    }
    return candidate;
};
var getFoldHandlerName = function (c) {
    return "on" + c.name;
};
var getFoldPositionalHandlers = function (d, isEager, usedConstructor) {
    var returnTypeParameterName = getFoldReturnTypeParameterName(d);
    return d.constructors.toArray().map(function (c) {
        var type = isEager && isNullaryConstructor(c)
            ? ts.createTypeReferenceNode(returnTypeParameterName, [])
            : ts.createFunctionTypeNode(undefined, c.members.map(function (m, position) { return getParameterDeclaration(getMemberName(m, position), getTypeNode(m.type)); }), ts.createTypeReferenceNode(returnTypeParameterName, []));
        var isParameterUnused = usedConstructor !== undefined && usedConstructor !== c;
        var foldHandlerName = isParameterUnused ? "_" + getFoldHandlerName(c) : getFoldHandlerName(c);
        return getParameterDeclaration(foldHandlerName, type);
    });
};
var getFoldRecordHandlers = function (d, handlersName, isEager) {
    var returnTypeParameterName = getFoldReturnTypeParameterName(d);
    var type = ts.createTypeLiteralNode(d.constructors.toArray().map(function (c) {
        var type = isEager && isNullaryConstructor(c)
            ? ts.createTypeReferenceNode(returnTypeParameterName, [])
            : ts.createFunctionTypeNode(undefined, c.members.map(function (m, position) { return getParameterDeclaration(getMemberName(m, position), getTypeNode(m.type)); }), ts.createTypeReferenceNode(returnTypeParameterName, []));
        return getPropertySignature(getFoldHandlerName(c), type, false);
    }));
    return [getParameterDeclaration(handlersName, type)];
};
var getFoldPositionalBody = function (d, matcheeName, tagName, isEager) {
    return ts.createBlock([
        ts.createSwitch(ts.createPropertyAccess(ts.createIdentifier(matcheeName), tagName), ts.createCaseBlock(d.constructors.toArray().map(function (c) {
            var access = ts.createIdentifier(getFoldHandlerName(c));
            return ts.createCaseClause(ts.createStringLiteral(c.name), [
                ts.createReturn(isEager && isNullaryConstructor(c)
                    ? access
                    : ts.createCall(access, [], c.members.map(function (m, position) {
                        return ts.createPropertyAccess(ts.createIdentifier(matcheeName), getMemberName(m, position));
                    })))
            ]);
        })))
    ]);
};
var getFoldRecordBody = function (d, matcheeName, tagName, handlersName, isEager) {
    return ts.createBlock([
        ts.createSwitch(ts.createPropertyAccess(ts.createIdentifier(matcheeName), tagName), ts.createCaseBlock(d.constructors.toArray().map(function (c) {
            var access = ts.createPropertyAccess(ts.createIdentifier(handlersName), getFoldHandlerName(c));
            return ts.createCaseClause(ts.createStringLiteral(c.name), [
                ts.createReturn(isEager && isNullaryConstructor(c)
                    ? access
                    : ts.createCall(access, [], c.members.map(function (m, position) {
                        return ts.createPropertyAccess(ts.createIdentifier(matcheeName), getMemberName(m, position));
                    })))
            ]);
        })))
    ]);
};
var getFold = function (d, name, isEager) {
    return new Reader_1.Reader(function (e) {
        var matcheeName = e.matcheeName;
        var tagName = e.tagName;
        var handlersStyle = e.handlersStyle;
        var returnTypeParameterName = getFoldReturnTypeParameterName(d);
        var typeParameters = d.parameterDeclarations
            .concat(M.parameterDeclaration(returnTypeParameterName, Option_1.none))
            .map(function (p) { return ts.createTypeParameterDeclaration(p.name, p.constraint.map(getTypeNode).toUndefined()); });
        var matchee = getParameterDeclaration(matcheeName, ts.createTypeReferenceNode(d.name, d.parameterDeclarations.map(function (p) { return ts.createTypeReferenceNode(p.name, []); })));
        var handlers = handlersStyle.type === 'positional'
            ? getFoldPositionalHandlers(d, isEager)
            : getFoldRecordHandlers(d, handlersStyle.handlersName, isEager);
        var parameters = [matchee].concat(handlers);
        var type = ts.createTypeReferenceNode(returnTypeParameterName, []);
        var body = handlersStyle.type === 'positional'
            ? getFoldPositionalBody(d, matcheeName, tagName, isEager)
            : getFoldRecordBody(d, matcheeName, tagName, handlersStyle.handlersName, isEager);
        return getFunctionDeclaration(name, typeParameters, parameters, type, body);
    });
};
exports.fold = function (d) {
    return Reader_1.ask().chain(function (e) {
        var folds = Array_1.empty;
        if (e.encoding === 'literal') {
            if (isSumType(d)) {
                if (isEagerFoldSupported(d)) {
                    folds = [getFold(d, e.foldName, true), getFold(d, e.foldName + "L", false)];
                }
                else {
                    folds = [getFold(d, e.foldName, false)];
                }
            }
        }
        return Array_1.array.sequence(Reader_1.reader)(folds);
    });
};
var isPolymorphic = function (d) {
    return d.parameterDeclarations.length > 0;
};
exports.prisms = function (d) {
    return Reader_1.ask().chain(function (e) {
        if (!isSumType(d)) {
            return Reader_1.reader.of(Array_1.empty);
        }
        var dataType = ts.createTypeReferenceNode(d.name, d.parameterDeclarations.map(function (p) { return ts.createTypeReferenceNode(p.name, []); }));
        var type = ts.createTypeReferenceNode('Prism', [dataType, dataType]);
        var getPrism = function (name) {
            return ts.createCall(ts.createPropertyAccess(ts.createIdentifier('Prism'), 'fromPredicate'), [], [
                ts.createArrowFunction([], [], [getParameterDeclaration('s')], undefined, undefined, ts.createBinary(ts.createPropertyAccess(ts.createIdentifier('s'), e.tagName), ts.SyntaxKind.EqualsEqualsEqualsToken, ts.createStringLiteral(name)))
            ]);
        };
        var monocleImport = ts.createImportDeclaration([], [], ts.createImportClause(undefined, ts.createNamedImports([ts.createImportSpecifier(undefined, ts.createIdentifier('Prism'))])), ts.createStringLiteral('monocle-ts'));
        var typeParameters = d.parameterDeclarations.map(function (p) {
            return ts.createTypeParameterDeclaration(p.name, p.constraint.map(getTypeNode).toUndefined());
        });
        if (isPolymorphic(d)) {
            return Reader_1.reader.of([
                monocleImport
            ].concat(d.constructors.toArray().map(function (c) {
                var body = ts.createBlock([ts.createReturn(getPrism(c.name))]);
                return getFunctionDeclaration("_" + getFirstLetterLowerCase(c.name), typeParameters, [], type, body);
            })));
        }
        return Reader_1.reader.of([
            monocleImport
        ].concat(d.constructors.toArray().map(function (c) {
            return getConstantDeclaration("_" + c.name, getPrism(c.name), type);
        })));
    });
};
