"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var A = require("./ast");
exports.ast = function (ast) {
    var printer = ts.createPrinter({
        newLine: ts.NewLineKind.LineFeed
    });
    var source = ts.createSourceFile('', '', ts.ScriptTarget.Latest);
    return printer.printNode(ts.EmitHint.Unspecified, ast, source);
};
exports.data = function (d) {
    return A.data(d).map(function (declarations) { return declarations.map(exports.ast).join('\n\n'); });
};
exports.constructors = function (d) {
    return A.constructors(d).map(function (nodes) { return nodes.map(exports.ast); });
};
exports.fold = function (d) {
    return A.fold(d).map(function (functionDeclarations) { return functionDeclarations.map(exports.ast); });
};
exports.prisms = function (d) {
    return A.prisms(d).map(function (nodes) { return nodes.map(exports.ast); });
};
exports.all = function (d) {
    return exports.data(d).chain(function (data) {
        return exports.constructors(d).chain(function (constructors) {
            return exports.fold(d).chain(function (folds) { return exports.prisms(d).map(function (prisms) { return [data].concat(constructors, folds, prisms); }); });
        });
    });
};
exports.print = function (d, options) {
    if (options === void 0) { options = A.defaultOptions; }
    return exports.all(d)
        .run(options)
        .join('\n\n');
};
